#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2021 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################


import wx
import time
import socket
import json
from pathlib import Path
from queue import Queue
import threading
from nardascripting.base.signalsharkdev import *
from nardascripting.base.usrscriptbase import *
from nardascripting.base.measthread import *
import nardascripting.base.toolbox as tb
from . import filehandling as fh


class DfNetSync(UsrScriptBase):
    """User script class"""

    def __init__(self, main_gui, dev=SignalSharkDev()):
        """Initialization. Please leave this line unchanged"""
        super().__init__(main_gui, dev, __file__)

        # Script settings.
        # -------------------------------------------------------------------------------
        # Please adapt the following lines of code according to your script.
        self._tab_name = "nslnetdf"
        self._scr_title = "Network device synchronization V1.1"
        self._scr_description = "Synchronizes all network SignalShark devices to the bearing settings " \
                                "of the master device"
        # self._icon_path = self.script_path.joinpath("NardaScriptLauncher_example01.png")
        self._list_prio = 2
        self._nsl_executed_behavior = NSL_Executed_Behaviors.SHOW_NSL

        # Add class variables if needed
        # -------------------------------------------------------------------------------
        self.device_list = {}  # Device list
        self.site_settings_list = {}  # Site settings list

    def _run_script(self, args):
        """
        Script main function
        This method is called when a user clicks on the corresponding script button.
        """
        # load device list:
        self.device_list = fh.load_from_json(fh.DEVICE_LIST_NAME)
        self.site_settings_list = fh.load_from_json(fh.DEVICE_SITE_SETTINGS_NAME)

        # Check if device list exists:
        if not self.device_list or not self.site_settings_list:
            self.MessageBoxModal("Please run script \'Scan Devices\' to generate device list.", "No device list",
                                 wx.OK | wx.ICON_INFORMATION)
            return

        mthread = MeasDlgThread(main_gui=self.main_gui, callback=DfNetSync._run_measurement,
                                args=[self.signalshark.addr, self.device_list, self.site_settings_list])
        mthread.start_measurement()

        if mthread.result is not None:
            self.MessageBoxModal(str(mthread.result), "Synchronization - Result",
                                 wx.OK | wx.ICON_INFORMATION)

    @staticmethod
    def _run_measurement(stopevent, update_progress, wait_for_stopevent,
                         master_ip: str, device_list: dict, site_settings_list: dict):
        """Main method of the measurement thread that handles a time consuming measurement.

        :param stopevent: Thread flag that indicates whether the procedure should be finished/cancelled
        :param update_progress: Delegate method to update a progress bar, message string and icon
        :param wait_for_stopevent: Delegate method to wait until stopevent is raised with possibility to show a message.
        :param master_ip: Custom parameter to handover the ip address of the SignalShark
        :param device_list: Custom parameter to handover the device list
        :param site_settings_list: Custom parameter to handover the isite settings list
        :return:
        """
        signalshark = SignalSharkDev(master_ip)
        try:
            if not signalshark.connect():
                return wait_for_stopevent("Cannot connect to master device!")

            master_pn = signalshark.scpi.device.get_part_number()
            master_sn = signalshark.scpi.device.get_serial_number()
            master_id = "{}_{}".format(master_pn, master_sn)


            task_type, task_name = signalshark.scpi.task.get_selected()
            if task_type != TaskTypes.AUTO_DF:
                return wait_for_stopevent("Please add Auto DF task to master and configure it!")

            # Read settings from master device:
            signalshark.scpi.bearing.set_data_valid_enable(True)
            bearing_signal_settings = signalshark.scpi.get_bearing_signal_settings()
            fspan = signalshark.scpi.spectrum.get_frequency_span()
            if bearing_signal_settings is None:
                return wait_for_stopevent("Cannot read settings from master device!")

            # Apply master bearing settings to network devices
            devices_failed = Queue()
            devices = Queue()
            for key in device_list:
                devices.put(key)

            num_threads = devices.qsize()
            if num_threads < 1:
                return wait_for_stopevent("Empty device list!")

            threads = []
            if num_threads > 5:
                num_threads = 5

            for x in range(num_threads):
                t = threading.Thread(target=DfNetSync.threader, args=(devices, master_id, bearing_signal_settings, fspan,
                                                                      site_settings_list, devices_failed,))
                t.daemon = True
                threads.append(t)
                t.start()

            # Wait until synchronization is finished and build return message
            devices.join()
            for t in threads:
                t.join()

            msg = "Synchronization done"
            num_failed = devices_failed.qsize()
            if num_failed > 0:
                msg.join("\r\nThe following devices could not be initialized:")
                for y in range(num_failed):
                    msg.join("\r\n" + devices_failed.get())
            return msg
        except Exception as e:
            # Do some error handling
            wait_for_stopevent(str(e))
            return "Error while synchronization!"
        finally:
            if signalshark is not None:
                signalshark.disconnect()

    @staticmethod
    def threader(devices: Queue, master_id: str, settings: BearingSignalSettings, fspan: float,
                 site_settings_list: dict, devices_failed: Queue):
        while not devices.empty():
            device_ip = devices.get(False)
            DfNetSync.apply_settings(device_ip, master_id, settings, fspan, site_settings_list, devices_failed)
            devices.task_done()

    @staticmethod
    def apply_settings(device_ip: str, master_id: str, settings: BearingSignalSettings, fspan: float,
                       site_settings_list: dict, devices_failed: Queue):
        error = False
        signalshark = SignalSharkDev(device_ip)
        try:
            signalshark.connect()
            if signalshark.scpi.check_no_error():
                device_pn = signalshark.scpi.device.get_part_number()
                device_sn = signalshark.scpi.device.get_serial_number()
                dev_id = "{}_{}".format(device_pn, device_sn)

                # Check if slave:
                if dev_id != master_id:
                    # Check if Auto DF task is set:
                    if signalshark.scpi.check_add_task(TaskTypes.AUTO_DF, "ADF_Script"):
                        # Apply bearing settings:
                        if not signalshark.scpi.set_bearing_signal_settings(settings):
                            error = True
                        signalshark.scpi.spectrum.set_frequency_span(fspan)

                # Apply site settings:
                dev_site_settings = site_settings_list[dev_id]
                site_settings = BearingSiteSettings()
                site_settings.attenuator = dev_site_settings["Atten"]
                site_settings.unit = LevelUnits.from_string(dev_site_settings["Unit"])
                site_settings.magnetic_declination = dev_site_settings["Decl"]
                site_settings.azimuth_correction = dev_site_settings["AzimCorr"]
                site_settings.reference_mark_direction = dev_site_settings["RefMarkDir"]
                site_settings.north_reference = BearingNorthRefs.from_string(dev_site_settings["NRef"])
                if not signalshark.scpi.set_bearing_site_settings(site_settings):
                    error = True

                auto_attn = dev_site_settings["AutoAtten"]
                if auto_attn:
                    if not DfNetSync.auto_attn_loop(signalshark, site_settings.attenuator):
                        error = True

        except:
            error = True
        finally:
            signalshark.disconnect()
            if error:
                devices_failed.put(device_ip)

    @staticmethod
    def auto_attn_loop(signalshark: SignalSharkDev, start_atten: float) -> bool:
        """Sets attenuator in that SignalShark is not overdriven.

        :param signalshark: The SignalShark device
        :param start_atten: The given start attenuation value
        :return: Returns True, if attenuation control loop was successful.
        """
        retvalue = False
        scan_cnt = 0
        try:
            atten_list = signalshark.scpi.sense.get_attenuator_list()
            max_atten = max(atten_list)
            atten_curr = start_atten
            valid_enabled_old = signalshark.scpi.bearing.get_data_valid_enable()
            signalshark.scpi.bearing.set_data_valid_enable(False)
            while True:
                overdr = False
                for x in range(5):
                    scan_cnt = signalshark.scpi.bearing.get_data_update()
                    bearingdata = signalshark.scpi.bearing.get_data_all()
                    overdr |= bearingdata.overdriven
                    if overdr:
                        break
                if overdr:
                    atten_curr += 3
                    if atten_curr > max_atten:
                        break
                    signalshark.scpi.sense.set_attenuator(atten_curr)
                else:
                    retvalue = True
                    break
            signalshark.scpi.bearing.set_data_valid_enable(valid_enabled_old)
        except:
            retvalue = False
        return retvalue
